package net.knarcraft.minecraftserverlauncher.profile;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import net.knarcraft.minecraftserverlauncher.userinterface.GUI;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerConsoles;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerTab;
import net.knarcraft.minecraftserverlauncher.server.AdvancedServerType;
import net.knarcraft.minecraftserverlauncher.server.Server;
import net.knarcraft.minecraftserverlauncher.server.ServerType;
import net.knarcraft.minecraftserverlauncher.Main;

import javax.swing.*;
import java.io.*;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.Scanner;
import java.util.concurrent.Executors;

import static net.knarcraft.minecraftserverlauncher.Shared.downloadFile;
import static net.knarcraft.minecraftserverlauncher.Shared.readFile;
import static net.knarcraft.minecraftserverlauncher.Shared.stringBetween;

/**
 * Contains all user settings, and a list of servers.
 *
 * @author      Kristian Knarvik <kristian.knarvik@knett.no>
 * @version     1.0.0
 * @since       1.0.0
 */
public class Profile {
    private static final ArrayList<Profile> profiles = new ArrayList<>();
    private static Profile current;
    private static GUI gui;
    private static final String profilesDir = Main.getApplicationWorkDirectory() + File.separator +  "files";
    private static final String profilesFile = Main.getApplicationWorkDirectory() + File.separator +  "files" + File.separator + "Profiles.txt";
    private static final String jarDirectory = Main.getApplicationWorkDirectory() + File.separator + "files" + File.separator + "Jars" + File.separator;

    private final ArrayList<Collection> collections;
    private final String name;
    private boolean runInBackground;
    private int delayStartup;
    private boolean downloadAllAvailableJARFiles;
    private static String vanillaVersion;
    private static String snapshotVersion;
    private static String bungeeVersion;

    private Profile(String name) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = false;
        this.delayStartup = 0;
        this.downloadAllAvailableJARFiles = false;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    private Profile(String name, boolean runInBackground, int delayStartup, boolean downloadAllAvailableJARFiles) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = runInBackground;
        this.delayStartup = delayStartup;
        this.downloadAllAvailableJARFiles = downloadAllAvailableJARFiles;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    public static GUI getGUI() {
        return gui;
    }

    public boolean getRunInBackground() {
        return this.runInBackground;
    }

    public int getDelayStartup() {
        return this.delayStartup;
    }

    public boolean getDownloadAllAvailableJARFiles() {
        return this.downloadAllAvailableJARFiles;
    }

    public static Profile getCurrent() {
        return current;
    }

    public ArrayList<Collection> getCollections() {
        return this.collections;
    }

    public String getName() {
        return this.name;
    }

    /**
     * Gets a Collection object by name.
     *
     * @param name  The name of the collection.
     * @return      A collection object.
     */
    public Collection getCollection(String name) {
        for (Collection collection : this.collections) {
            if (collection.getName().equals(name)) {
                return collection;
            }
        }
        return null;
    }

    public static Profile getProfile(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                return profile;
            }
        }
        return null;
    }

    public static ArrayList<Profile> getProfiles() {
        return profiles;
    }

    public void setRunInBackground(boolean value) {
        this.runInBackground = value;
    }

    public void setDelayStartup(int value) {
        if (value >= 0) {
            this.delayStartup = value;
        }
    }

    public void setDownloadAllAvailableJARFiles(boolean value) {
        this.downloadAllAvailableJARFiles = value;
    }

    /**
     * Set the current profile to the profile with a certain name.
     *
     * @param name  The name of the profile
     */
    public static  void setCurrent(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                current = profile;
                break;
            }
        }
    }

    /**
     * Adds a collection to the profile if the name is valid.
     *
     * @param name  The name of the collection and its elements.
     */
    public void addCollection(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (getCollection(name) == null &&
                !name.equals("") &&
                !name.equals("All") &&
                !name.contains("!") &&
                !name.contains("?") &&
                !name.contains(";")
                ) {
            collections.add(new Collection(name));
        } else {
            showError("A server name must my unique and not empty or \"All\"." +
                    "It can't contain any of the characters \"!\", \"?\" or \";\".");
        }
    }

    /**
     * Adds a profile if the name is valid and unique.
     *
     * @param name  The name of the new profile.
     */
    public static void addProfile(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (name.equals("") && !name.matches("^[!?;]+$")) {
            showError("Profile name can't be blank.");
            return;
        }
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                showError("There is already a profile with this name.");
                return;
            }
        }
        new Profile(name);
    }

    public void removeCollection(String name) {
        for (int i = 0; i < collections.size(); i++) {
            if (collections.get(i).getName().equals(name)) {
                this.collections.remove(i);
                gui.update();
                break;
            }
        }
    }

    /**
     * Removes a profile with the given name from the list of profiles, if such a profile exists
     * @param name <p>The name of the profile to rempve</p>
     */
    public static void removeProfile(String name) {
        if (profiles.size() > 1) {
            profiles.removeIf(profile -> profile.name.equals(name));
        }
    }

    public void updateConsoles() {
        JTabbedPane consolesTab = ServerConsoles.getTabbedPane();
        consolesTab.removeAll();
        for (Collection collection : collections) {
            consolesTab.add(collection.getName(), collection.getServerConsole().getPanel());
        }
    }

    /**
     * Sends a command to a server, or all servers
     *
     * @param serverName    The target server
     * @param command       The command to send.
     */
    public void sendCommand(String serverName, String command) {
        if (serverName.equals("All")) {
            for (Collection collection : this.collections) {
                try {
                    collection.getServer().sendCommand(command);
                } catch (IOException e) {
                    showError("Server " + collection.getName() + " caused an exception.");
                }
            }
        } else {
            Collection collection = getCollection(serverName);
            if (collection != null) {
                Server target = collection.getServer();
                try {
                    target.sendCommand(command);
                } catch (IOException e) {
                    showError("Server " + target.getName() + " caused an exception.");
                }
            } else {
                showError("Server " + serverName + " is invalid.");
            }
        }
    }

    /**
     * Reads all server tabs, and saves it to the variables of the corresponding servers.
     * Saves all profiles and servers to a text file.
     */
    public void save() throws FileNotFoundException {
        for (Collection collection : this.collections) {
            Server server = collection.getServer();
            ServerTab serverTab = collection.getServerTab();
            server.setPath(serverTab.getPath());
            server.setMaxRam(serverTab.getMaxRam());
            server.setType(ServerType.getByName(serverTab.getType()));
            try {
                server.setServerVersion(serverTab.getVersion());
            } catch (IllegalArgumentException e) {
                showError("Invalid server version for " + server.getName());
            }
            server.toggle(serverTab.enabled());
        }
        if (!new File(profilesDir).exists() && !new File(profilesDir).mkdirs()) {
            showError("Unable to create the folder " + profilesDir);
            throw new FileNotFoundException("Unable to create the profiles folder: " + profilesDir);
        }
        try (PrintWriter file = new PrintWriter(profilesFile)) {
            int width;
            int height;
            if (gui == null) {
                width = 440;
                height = 170;
            } else {
                width = gui.getSize().width;
                height = gui.getSize().height;
            }
            file.println(String.format(
                    "%s;%s;%s;%s;%d;%d",
                    current.name,
                    vanillaVersion,
                    snapshotVersion,
                    bungeeVersion,
                    width,
                    height
            ));
            file.close();
            for (Profile profile : profiles) {
                StringBuilder saveString = new StringBuilder(String.format(
                        "%s;%b;%d;%b?",
                        profile.name,
                        profile.runInBackground,
                        profile.delayStartup,
                        profile.downloadAllAvailableJARFiles)
                );
                for (Collection collection : profile.collections) {
                    Server server = collection.getServer();
                    saveString.append(String.format(
                            "%s;%s;%b;%s;%s;%s;%s;%s;%s;%s!",
                            server.getName(),
                            server.getPath(),
                            server.isEnabled(),
                            server.getTypeName(),
                            server.getServerVersion(),
                            server.getMaxRam(),
                            server.getVanillaVersion(),
                            server.getSnapshotVersion(),
                            server.getSpongeVanillaVersion(),
                            server.getBungeeVersion()
                            )
                    );
                }
                saveString = new StringBuilder(saveString.substring(0, saveString.length() - 1));
                try (PrintWriter fileAppend = new PrintWriter(new FileWriter(
                        profilesFile,
                        true
                ))) {
                    fileAppend.println(saveString);
                } catch (IOException e) {
                    if (gui != null) {
                        showError("Unable to save to file. Try running the software as an administrator.");
                    } else {
                        System.out.println("Unable to save to file. Try running the software as an administrator.");
                    }
                    throw new FileNotFoundException("Unable to save to the profiles file.");
                }
            }
        } catch (IOException e) {
            if (gui != null) {
                showError("Unable to save to file. Try running the software as an administrator.");
            }
            throw new FileNotFoundException("Unable to create the profiles file");
        }
    }

    /**
     * Reads profiles and servers from a text file.
     */
    public static void load() {
        try (Scanner in = new Scanner(new File(profilesFile))) {
            try {
                String[] staticData = in.nextLine().split(";", -1);
                String profileName = staticData[0];
                vanillaVersion = staticData[1];
                snapshotVersion = staticData[2];
                bungeeVersion = staticData[3];
                int guiWidth = Integer.parseInt(staticData[4]);
                int guiHeight = Integer.parseInt(staticData[5]);
                gui = new GUI(guiWidth, guiHeight);
                while (in.hasNextLine()) {
                    String line = in.nextLine();
                    if (line.contains("?")) {
                        String[] data = line.split("\\?");
                        String[] profileData = data[0].split(";", -1);
                        Profile profile = parseProfile(profileData);
                        if (data[1].contains("!")) {
                            String[] servers = data[1].split("!", -1);
                            for (String server : servers) {
                                String[] serverData = server.split(";", -1);
                                parseServer(profile, serverData);
                            }
                        } else {
                            String[] serverData = data[1].split(";", -1);
                            parseServer(profile, serverData);
                        }
                    } else {
                        String[] profileData = line.split(";", -1);
                        parseProfile(profileData);
                    }
                }
                current = getProfile(profileName);
            } catch (ArrayIndexOutOfBoundsException | NumberFormatException e) {
                e.printStackTrace();
                showError("Invalid Profile.txt file. Profiles could not be loaded. If this error persists, please " +
                        "manually delete the file.");
                System.exit(1);
            } catch (IOException e) {
                e.printStackTrace();
            }
            if (profiles.size() == 0) {
                addProfile("Default");
            }
        } catch (FileNotFoundException | NoSuchElementException e) {
            showMessage("A profiles file was not found. Default profile was created.");
            try {
                gui = new GUI();
            } catch (FileNotFoundException ex) {
                showMessage("Failed to load GUI messages. The GUI can't be shown.");
            } catch (IOException ex) {
                ex.printStackTrace();
            }
            addProfile("Default");
        }
        gui.update();
        gui.updateProfiles();
        current.updateConsoles();
        if (current.downloadAllAvailableJARFiles) {
            Executors.newSingleThreadExecutor().execute(() -> {
                try {
                    downloadJars();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            });
        }
        if (current.runInBackground) {
            Executors.newSingleThreadExecutor().execute(Server::startServers);
            gui.hide();
        }
    }

    /**
     * Parses a profile, and creates a profile with the data.
     * @param profileData <p>The data of the new profile</p>
     * @return <p>The new profile</p>
     */
    private static Profile parseProfile(String[] profileData) {
        return new Profile(
                profileData[0],
                Boolean.parseBoolean(profileData[1]),
                Integer.parseInt(profileData[2]),
                Boolean.parseBoolean(profileData[3])
        );
    }

    /**
     * Parses a server, and creates a new collection.
     * @param profile <p>The profile which to add the collection</p>
     * @param serverData <p>The data to parse</p>
     */
    private static void parseServer(Profile profile, String[] serverData) {
        profile.collections.add(new Collection(
                serverData[0],
                serverData[1],
                Boolean.parseBoolean(serverData[2]),
                serverData[3],
                serverData[4],
                serverData[5],
                serverData[6],
                serverData[7],
                serverData[8],
                serverData[9])
        );
    }

    /**
     * Downloads all jars to the program directory.
     *
     * @throws IOException  On version file failure or folder creation failure
     */
    public static void downloadJars() throws IOException {
        if (!new File(jarDirectory).exists() && !new File(jarDirectory).mkdirs()) {
            showError("Could not create the Jars folder. Please run the program with admin permissions, or move it to " +
                    "a writable directory.");
            throw new FileNotFoundException("Unable to create jars folder");
        }
        try {
            downloadAll();
            printToGui("Finished downloading jars");
        } catch (FileNotFoundException e) {
            throw new FileNotFoundException("One or more downloads failed: " + e.getMessage());
        }
    }

    /**
     * Prints something to the gui status field if the gui exists
     * Otherwise it prints to the console
     * @param message <p>The string to show the user</p>
     */
    private static void printToGui(String message) {
        if (gui != null) {
            gui.setStatus(message);
        } else {
            System.out.println(message);
        }
    }

    public static void showError(String title, String message) {
        if (gui != null) {
            JOptionPane.showMessageDialog(
                    null,
                    message,
                    title,
                    JOptionPane.ERROR_MESSAGE
            );
        } else {
            System.out.println(message);
        }
    }

    public static void showError(String message) {
        if (gui != null) {
            JOptionPane.showMessageDialog(
                    null,
                    message,
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
        } else {
            System.out.println(message);
        }
    }

    public static void showMessage(String title, String message) {
        if (gui != null) {
            JOptionPane.showMessageDialog(
                    null,
                    message,
                    title,
                    JOptionPane.INFORMATION_MESSAGE
            );
        } else {
            System.out.println(message);
        }
    }

    public static void showMessage(String message) {
        if (gui != null) {
            JOptionPane.showMessageDialog(
                null,
                message,
                "Info",
                JOptionPane.INFORMATION_MESSAGE
            );
        } else {
            System.out.println(message);
        }
    }

    private static void showMessage() {}

    /**
     * Downloads jar files for all possible server versions
     * @throws IOException <p>If a jar fails to download</p>
     */
    private static void downloadAll() throws IOException {
        for (ServerType type : ServerType.getServerTypes()) {
            String downloadURL = Objects.requireNonNull(type).getDownloadURL();
            String typeName = type.getName();
            AdvancedServerType advancedServerType = type instanceof AdvancedServerType ? (AdvancedServerType) type : null;
            for (String version : type.getVersions()) {
                boolean success;
                printToGui("Downloading: " + typeName + version + ".jar");
                File file = new File(jarDirectory + type.getName() + version + ".jar");
                Path filePath = Paths.get(jarDirectory + type.getName() + version + ".jar");
                switch (type.getName()) {
                    case "Vanilla":
                    case "Snapshot":
                        success = downloadVanillaJar(advancedServerType, file, downloadURL, filePath, typeName, version);
                        break;
                    case "Spigot":
                    case "Craftbukkit":
                    case "MCPCplus":
                        success = downloadSpigotJar(file, downloadURL, typeName, version, filePath);
                        break;
                    case "SpongeVanilla":
                        success = downloadSpongeVanillaJar(advancedServerType, file, downloadURL, filePath, version);
                        break;
                    case "Bungee":
                        success = downloadBungeeJar(advancedServerType, file, downloadURL, filePath, typeName);
                        break;
                    default:
                        success = true;
                }
                if (!success) {
                    printToGui("Error downloading: " + typeName + version + ".jar");
                    throw new FileNotFoundException("Error downloading: " + typeName + version + ".jar");
                }
            }
        }
    }

    /**
     * Downloads a Spigot, Craftbukkit or MCPC+ .jar file
     * @param file <p>The file the .jar file should be saved as</p>
     * @param downloadURL <p>The base URL for downloading the .jar file</p>
     * @param typeName <p>The name of the selected server type</p>
     * @param version <p>The version of the .jar file to download</p>
     * @param filePath <p>The path of the .jar file</p>
     * @return <p>True if the file exists or the file was successfully downloaded</p>
     */
    private static boolean downloadSpigotJar(File file, String downloadURL, String typeName, String version, Path filePath) {
        return file.isFile() || downloadFile(downloadURL + typeName + version + ".jar", filePath);
    }

    /**
     * Downloads a Vanilla/Snapshot .jar file according to the input
     * @param advancedServerType <p>The advanced server type containing required information</p>
     * @param file <p>The file the .jar file should be saved as</p>
     * @param downloadURL <p>The base URL for downloading the .jar file</p>
     * @param filePath <p>The path of the .jar file</p>
     * @param typeName <p>The name of the selected server type</p>
     * @param version <p>The version of the .jar file to download</p>
     * @return <p>True if the file exists or the file was successfully downloaded</p>
     * @throws IOException <p>If something goes horribly wrong</p>
     */
    private static boolean downloadVanillaJar(AdvancedServerType advancedServerType, File file, String downloadURL,
                                              Path filePath, String typeName, String version) throws IOException {
        if (version.equals("Latest")) {
            String versionText = readFile(Objects.requireNonNull(advancedServerType).getVersionURL());
            JsonObject jsonObject = new JsonParser().parse(versionText).getAsJsonObject();
            String latest = jsonObject.getAsJsonObject("latest").get("release").getAsString();
            JsonElement ver = jsonObject.getAsJsonArray("versions").get(0);
            String versionFile = ver.getAsJsonObject().get("url").getAsString();
            versionText = readFile(versionFile);
            jsonObject = new JsonParser().parse(versionText).getAsJsonObject();
            String jarFile = jsonObject.getAsJsonObject("downloads").getAsJsonObject("server").get("url").getAsString();
            setVersion(typeName, latest);
            return (file.isFile() && latest.equals(getVersion(typeName))) || downloadFile(jarFile, filePath);
        } else {
            return file.isFile() || downloadFile(downloadURL + version +
                    Objects.requireNonNull(advancedServerType).getDownloadURLPart() + version + ".jar", filePath);
        }
    }

    /**
     * Downloads a SpongeVanilla .jar file according to the input
     * @param advancedServerType <p>The advanced server type containing required information</p>
     * @param file <p>The file the .jar file should be saved as</p>
     * @param downloadURL <p>The base URL for downloading the .jar file</p>
     * @param filePath <p>The path of the .jar file</p>
     * @param version <p>The version of the .jar file to download</p>
     * @return <p>True if the file exists or the file was successfully downloaded</p>
     * @throws IOException <p>If something goes horribly wrong</p>
     */
    private static boolean downloadSpongeVanillaJar(AdvancedServerType advancedServerType, File file,
                                                    String downloadURL, Path filePath,
                                                    String version) throws IOException {
        String newestVersion = stringBetween(readFile(Objects.requireNonNull(advancedServerType).getVersionURL()
                + version), advancedServerType.getSrcStart(), advancedServerType.getSrcEnd());
        return file.isFile() || downloadFile(downloadURL + newestVersion + advancedServerType.getDownloadURLPart()
                +  newestVersion + ".jar", filePath);
    }

    /**
     * Downloads a Bungee .jar file according to the input
     * @param advancedServerType <p>The advanced server type containing required information</p>
     * @param file <p>The file the .jar file should be saved as</p>
     * @param downloadURL <p>The base URL for downloading the .jar file</p>
     * @param filePath <p>The path of the .jar file</p>
     * @param typeName <p>The name of the selected server type</p>
     * @return <p>True if the file exists or the file was successfully downloaded</p>
     * @throws IOException <p>If something goes horribly wrong</p>
     */
    private static boolean downloadBungeeJar(AdvancedServerType advancedServerType, File file,
            String downloadURL, Path filePath, String typeName) throws IOException {
        String newestVersion = stringBetween(readFile(Objects.requireNonNull(advancedServerType).getVersionURL()),
                advancedServerType.getSrcStart(), advancedServerType.getSrcEnd());
        setVersion(typeName, newestVersion);
        return (file.isFile() && newestVersion.equals(getVersion(typeName))) || downloadFile(downloadURL, filePath);
    }

    /**
     * Returns the current version of a type
     * @param type <p>The version type</p>
     * @return <p>The version string</p>
     */
    private static String getVersion(String type) {
        switch (type) {
            case "Vanilla":
                return vanillaVersion;
            case "Snapshot":
                return snapshotVersion;
            case "Bungee":
                return bungeeVersion;
            default:
                return "";
        }
    }

    /**
     * Sets a server type's last downloaded version
     * @param type <p>The version type</p>
     * @param version <p>The version string</p>
     */
    private static void setVersion(String type, String version) {
        if (!type.equals("")) {
            switch (type) {
                case "Vanilla":
                    vanillaVersion = version;
                    break;
                case "Snapshot":
                    snapshotVersion = version;
                    break;
                case "Bungee":
                    bungeeVersion = version;
            }
        }
    }
}
