package net.knarcraft.minecraftserverlauncher.server;

import javax.naming.ConfigurationException;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Scanner;

import static net.knarcraft.minecraftserverlauncher.Shared.getResourceAsScanner;

/**
 * Contains the bare minimum to be a functional server type.
 *
 * @author      Kristian Knarvik <kristian.knarvik@knett.no>
 * @version     1.0.0
 * @since       1.0.0
 */
public class ServerType {
	private final String name;
	private final String[] versions;
	private final String downloadURL;
	private static final ArrayList<ServerType> serverTypes = new ArrayList<>();

	/**
	 * Instantiates a new server type
	 * @param name <p>The name of the server type</p>
	 * @param versions <p>A list of one or more server versions for the type</p>
	 * @param downloadURL <p>The URL used for downloading .jar files</p>
	 */
	ServerType(String name, String[] versions, String downloadURL) {
		this.name = name;
		this.versions = versions;
		this.downloadURL = downloadURL;
		serverTypes.add(this);
	}

	/**
	 * Gets the name of the server type
	 * @return <p>Server type name</p>
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Gets a list of versions available for the server type
	 * @return <p>A list of server versions</p>
	 */
	public String[] getVersions() {
		return this.versions;
	}

	/**
	 * Gets the url used for downloading JAR files
	 * @return <p>A download URL</p>
	 */
	public String getDownloadURL() {
		return this.downloadURL;
	}

	/**
	 * Gets all instantiated server types
	 * @return <p>A list of server types</p>
	 */
	public static ArrayList<ServerType> getServerTypes() {
		return serverTypes;
	}

	/**
	 * Gets a list of all server types' names.
	 * @return	<p>A list of strings</p>
	 */
	public static String[] getTypeNames() {
		ArrayList<ServerType> types = ServerType.getServerTypes();
		String[] serverTypeNames = new String[types.size()];
		for (int i = 0; i < types.size(); i++) {
			serverTypeNames[i] = types.get(i).getName();
		}
		return serverTypeNames;
	}

	/**
	 * Gets a server type by the given name
	 * @param name <p>Then name of the server type</p>
	 * @return <p>A ServerType</p>
	 */
	public static ServerType getByName(String name) {
		for (ServerType serverType : serverTypes) {
			if (serverType.getName().equals(name)) {
				return serverType;
			}
		}
		return null;
	}

	/**
	 * Reads valid server types and version from a file, and creates their objects.
	 * @throws ConfigurationException <p>If anything goes wrong</p>
	 */
	public static void loadServerTypes() throws ConfigurationException {
		if (serverTypes.isEmpty()) {
			Scanner file;
			try {
				file = getResourceAsScanner("servertypes.csv");
			} catch (FileNotFoundException e) {
				throw new ConfigurationException("Server type configuration file is missing.");
			}
			while (file.hasNextLine()) {
				//Splits the next file line into arguments
                String[] fileLine = file.nextLine().split(";", -1);
                int lineLength = fileLine.length;
                //Gets list of server versions from file line
                String[] serverVersion;
                if (fileLine[1].contains(",")) {
                    serverVersion = fileLine[1].split(",", -1);
                } else {
                    serverVersion = new String[]{fileLine[1]};
                }
                switch (lineLength) {
                    case 7:
                        new AdvancedServerType(fileLine[0], serverVersion, fileLine[2], fileLine[3], fileLine[4], fileLine[5], fileLine[6]);
                        break;
                    case 3:
                        new ServerType(fileLine[0], serverVersion, fileLine[2]);
                        break;
                    default:
                        throw new ConfigurationException("Error: Configuration file invalid.");
                }
            }
		} else {
			throw new ConfigurationException("Error: Configuration already loaded.");
		}
	}
}